@@ -73,6 +73,7 @@ extern "C" {
 
 #if MYSQL_VERSION_ID >= 50500
 #  define MRN_HANDLER_HAVE_TRUNCATE
+#  define MRN_HANDLER_HAVE_GET_PARENT_FOREIGN_KEY_LIST
 #endif
 
 #if MYSQL_VERSION_ID >= 50500
@@ -106,6 +107,22 @@ extern "C" {
 #  define MRN_HAVE_HA_EXTRA_PREPARE_FOR_FORCED_CLOSE
 #endif
 
+#if MYSQL_VERSION_ID >= 50604
+#  define MRN_TIMESTAMP_USE_TIMEVAL
+#elif defined(MRN_MARIADB_P)
+#  define MRN_TIMESTAMP_USE_MY_TIME_T
+#else
+#  define MRN_TIMESTAMP_USE_LONG
+#endif
+
+#if MYSQL_VERSION_ID < 50600 && !defined(MRN_MARIADB_P)
+#  define MRN_FIELD_STORE_TIME_NEED_TYPE
+#endif
+
+#if MYSQL_VERSION_ID < 50500
+#  define MRN_HAVE_TL_WRITE_ALLOW_READ
+#endif
+
 class ha_mroonga;
 
 /* structs */
@@ -142,6 +159,9 @@ public:
   MEM_ROOT  *mem_root_for_clone;
   grn_obj   key_buffer;
   grn_id    record_id;
+  grn_id    *key_id;
+  grn_id    *del_key_id;
+  MY_BITMAP multiple_column_key_bitmap;
 
 private:
   grn_ctx *ctx;
@@ -169,8 +189,10 @@ private:
   grn_table_cursor *cursor;
   grn_table_cursor *index_table_cursor;
   grn_obj *score_column;
+  grn_obj *id_accessor;
   grn_obj *key_accessor;
 
+  grn_obj *sorted_result;
   grn_obj *matched_record_keys;
 
   uchar **key_min;
@@ -185,6 +207,7 @@ private:
   bool fast_order_limit_with_index;
 
   bool ignoring_duplicated_key;
+  bool inserting_with_update;
   bool fulltext_searching;
 
 #ifdef MRN_HANDLER_HAVE_FINAL_ADD_INDEX
@@ -233,10 +256,11 @@ public:
   int update_row(const uchar *old_data, uchar *new_data);
   int delete_row(const uchar *buf);
 
-  uint max_supported_record_length() const { return HA_MAX_REC_LENGTH; }
-  uint max_supported_keys()          const { return 100; }
+  uint max_supported_record_length()   const { return HA_MAX_REC_LENGTH; }
+  uint max_supported_keys()            const { return 100; }
   uint max_supported_key_parts();
-  uint max_supported_key_length()    const { return MAX_KEY_LENGTH; }
+  uint max_supported_key_length()      const { return MAX_KEY_LENGTH; }
+  uint max_supported_key_part_length() const { return MAX_KEY_LENGTH; }
 
   ha_rows records_in_range(uint inx, key_range *min_key, key_range *max_key);
   int index_init(uint idx, bool sorted);
@@ -350,6 +374,9 @@ public:
   void restore_auto_increment(ulonglong prev_insert_id);
   void release_auto_increment();
   int reset_auto_increment(ulonglong value);
+  bool was_semi_consistent_read();
+  void try_semi_consistent_read(bool yes);
+  void unlock_row();
   int start_stmt(THD *thd, thr_lock_type lock_type);
 
 protected:
@@ -378,6 +405,17 @@ protected:
 #endif
   void change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
   bool primary_key_is_clustered();
+  bool is_fk_defined_on_table_or_index(uint index);
+  char *get_foreign_key_create_info();
+  char *get_tablespace_name(THD *thd, char *name, uint name_len);
+  bool can_switch_engines();
+  int get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+#ifdef MRN_HANDLER_HAVE_GET_PARENT_FOREIGN_KEY_LIST
+  int get_parent_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+#endif
+  uint referenced_by_foreign_key();
+  void init_table_handle_for_HANDLER();
+  void free_foreign_key_create_info(char* str);
 
 private:
   void push_warning_unsupported_spatial_index_search(enum ha_rkey_function flag);
@@ -386,6 +424,7 @@ private:
   void clear_search_result();
   void clear_search_result_geo();
   void clear_indexes();
+  int alter_share_add(const char *path, TABLE_SHARE *table_share);
   void remove_grn_obj_force(const char *name);
   grn_obj *find_tokenizer(const char *name, int name_length);
   int wrapper_get_next_record(uchar *buf);
@@ -396,12 +435,21 @@ private:
 #ifdef MRN_HANDLER_HAVE_HA_CLOSE
   int close();
 #endif
+  bool is_dry_write();
+  bool is_enable_optimization();
+  bool is_need_normalize(Field *field) const;
+  bool is_temporary_table_name(const char *name) const;
   void check_count_skip(key_part_map start_key_part_map,
                         key_part_map end_key_part_map, bool fulltext);
+  bool is_groonga_layer_condition(const Item *item,
+                                  const Item_func **match_against=NULL);
+  bool is_fulltext_search_item(const Item *item);
   void check_fast_order_limit(grn_table_sort_key **sort_keys, int *n_sort_keys,
                               longlong *limit,
                               grn_obj *target_table, grn_obj *score_column);
 
+  long long int get_grn_time_from_timestamp_field(Field_timestamp *field);
+
   int generic_store_bulk_string(Field *field, grn_obj *buf);
   int generic_store_bulk_integer(Field *field, grn_obj *buf);
   int generic_store_bulk_float(Field *field, grn_obj *buf);
@@ -409,7 +457,14 @@ private:
   int generic_store_bulk_date(Field *field, grn_obj *buf);
   int generic_store_bulk_time(Field *field, grn_obj *buf);
   int generic_store_bulk_datetime(Field *field, grn_obj *buf);
+  int generic_store_bulk_year(Field *field, grn_obj *buf);
   int generic_store_bulk_new_date(Field *field, grn_obj *buf);
+#ifdef MRN_HAVE_MYSQL_TYPE_DATETIME2
+  int generic_store_bulk_datetime2(Field *field, grn_obj *buf);
+#endif
+#ifdef MRN_HAVE_MYSQL_TYPE_TIME2
+  int generic_store_bulk_time2(Field *field, grn_obj *buf);
+#endif
   int generic_store_bulk_new_decimal(Field *field, grn_obj *buf);
   int generic_store_bulk_blob(Field *field, grn_obj *buf);
   int generic_store_bulk_geometry(Field *field, grn_obj *buf);
@@ -429,17 +484,42 @@ private:
                                 const char *value, uint value_length);
   void storage_store_field_datetime(Field *field,
                                     const char *value, uint value_length);
+  void storage_store_field_year(Field *field,
+                                const char *value, uint value_length);
   void storage_store_field_new_date(Field *field,
                                     const char *value, uint value_length);
+#ifdef MRN_HAVE_MYSQL_TYPE_DATETIME2
+  void storage_store_field_datetime2(Field *field,
+                                     const char *value, uint value_length);
+#endif
+#ifdef MRN_HAVE_MYSQL_TYPE_TIME2
+  void storage_store_field_time2(Field *field,
+                                 const char *value, uint value_length);
+#endif
   void storage_store_field_blob(Field *field,
                                 const char *value, uint value_length);
   void storage_store_field_geometry(Field *field,
                                     const char *value, uint value_length);
   void storage_store_field(Field *field, const char *value, uint value_length);
   void storage_store_fields(uchar *buf, grn_id record_id);
+  void storage_store_fields_for_prep_update(const uchar *old_data,
+                                            uchar *new_data,
+                                            grn_id record_id);
   void storage_store_fields_by_index(uchar *buf);
 
+  int storage_encode_key_time(Field *field, const uchar *key,
+                              uchar *buf, uint *size);
+  int storage_encode_key_year(Field *field, const uchar *key,
+                              uchar *buf, uint *size);
+#ifdef MRN_HAVE_MYSQL_TYPE_TIME2
+  int storage_encode_key_time2(Field *field, const uchar *key,
+                               uchar *buf, uint *size);
+#endif
   int storage_encode_key(Field *field, const uchar *key, uchar *buf, uint *size);
+  int storage_encode_multiple_column_key(KEY *key_info,
+                                         const uchar *key, uint key_length,
+                                         uchar *buffer, uint *encoded_length,
+                                         bool decode);
 
   void set_pk_bitmap();
   int wrapper_create(const char *name, TABLE *table,
@@ -447,20 +527,20 @@ private:
   int storage_create(const char *name, TABLE *table,
                      HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
   int wrapper_create_index_fulltext_validate(KEY *key_info);
-  int wrapper_create_index_fulltext(grn_obj *grn_table,
-                                    const char *grn_table_name,
+  int wrapper_create_index_fulltext(const char *grn_table_name,
                                     int i,
                                     KEY *key_info,
                                     grn_obj **index_tables,
                                     MRN_SHARE *tmp_share);
-  int wrapper_create_index_geo(grn_obj *grn_table,
-                               const char *grn_table_name,
+  int wrapper_create_index_geo(const char *grn_table_name,
                                int i,
                                KEY *key_info,
                                grn_obj **index_tables,
                                MRN_SHARE *tmp_share);
   int wrapper_create_index(const char *name, TABLE *table,
-                           HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
+                           HA_CREATE_INFO *info, MRN_SHARE *tmp_share,
+                           char *grn_table_name);
+  int wrapper_create_index_check(handler *hnd, TABLE *table);
   int storage_create_validate_pseudo_column(TABLE *table);
   int storage_create_validate_index(TABLE *table);
   int storage_create_index(TABLE *table, const char *grn_table_name,
@@ -487,7 +567,7 @@ private:
   void wrapper_overwrite_index_bits();
   int wrapper_close();
   int storage_close();
-  int mrn_extra(enum ha_extra_function operation);
+  int generic_extra(enum ha_extra_function operation);
   int wrapper_extra(enum ha_extra_function operation);
   int storage_extra(enum ha_extra_function operation);
   int wrapper_extra_opt(enum ha_extra_function operation, ulong cache_size);
@@ -514,15 +594,30 @@ private:
   int storage_write_row_index(uchar *buf, grn_id record_id,
                               KEY *key_info, grn_obj *index_column);
   int storage_write_row_indexes(uchar *buf, grn_id record_id);
+  int storage_write_row_unique_index(uchar *buf, grn_id record_id,
+                                     KEY *key_info, grn_obj *index_table,
+                                     grn_id *key_id);
+  int storage_write_row_unique_indexes(uchar *buf, grn_id record_id);
   int wrapper_get_record_id(uchar *data, grn_id *record_id, const char *context);
   int wrapper_update_row(const uchar *old_data, uchar *new_data);
   int wrapper_update_row_index(const uchar *old_data, uchar *new_data);
   int storage_update_row(const uchar *old_data, uchar *new_data);
   int storage_update_row_index(const uchar *old_data, uchar *new_data);
+  int storage_update_row_unique_indexes(uchar *new_data, grn_id record_id);
   int wrapper_delete_row(const uchar *buf);
   int wrapper_delete_row_index(const uchar *buf);
   int storage_delete_row(const uchar *buf);
   int storage_delete_row_index(const uchar *buf);
+  int storage_delete_row_unique_index(grn_obj *index_table, grn_id del_key_id);
+  int storage_delete_row_unique_indexes();
+  int storage_prepare_delete_row_unique_index(const uchar *buf,
+                                              grn_id record_id,
+                                              KEY *key_info,
+                                              grn_obj *index_table,
+                                              grn_obj *index_column,
+                                              grn_id *del_key_id);
+  int storage_prepare_delete_row_unique_indexes(const uchar *buf,
+                                                grn_id record_id);
   uint wrapper_max_supported_key_parts();
   uint storage_max_supported_key_parts();
   ulonglong wrapper_table_flags() const;
@@ -589,6 +684,8 @@ private:
   void storage_ft_end();
   FT_INFO *wrapper_ft_init_ext(uint flags, uint key_nr, String *key);
   FT_INFO *storage_ft_init_ext(uint flags, uint key_nr, String *key);
+  void generic_ft_init_ext_add_conditions_fast_order_limit(
+      struct st_mrn_ft_info *info, grn_obj *expression);
   FT_INFO *generic_ft_init_ext(uint flags, uint key_nr, String *key);
   int wrapper_ft_read(uchar *buf);
   int storage_ft_read(uchar *buf);
@@ -723,6 +820,7 @@ private:
   int storage_add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys);
 #endif
   int storage_add_index_multiple_columns(KEY *key_info, uint num_of_keys,
+                                         grn_obj **index_tables,
                                          grn_obj **index_columns);
 #ifdef MRN_HANDLER_HAVE_FINAL_ADD_INDEX
   int wrapper_final_add_index(handler_add_index *add, bool commit);
@@ -752,12 +850,38 @@ private:
   void storage_release_auto_increment();
   int wrapper_reset_auto_increment(ulonglong value);
   int storage_reset_auto_increment(ulonglong value);
+  bool wrapper_was_semi_consistent_read();
+  bool storage_was_semi_consistent_read();
+  void wrapper_try_semi_consistent_read(bool yes);
+  void storage_try_semi_consistent_read(bool yes);
+  void wrapper_unlock_row();
+  void storage_unlock_row();
   int wrapper_start_stmt(THD *thd, thr_lock_type lock_type);
   int storage_start_stmt(THD *thd, thr_lock_type lock_type);
   void wrapper_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
   void storage_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
   bool wrapper_primary_key_is_clustered();
   bool storage_primary_key_is_clustered();
+  bool wrapper_is_fk_defined_on_table_or_index(uint index);
+  bool storage_is_fk_defined_on_table_or_index(uint index);
+  char *wrapper_get_foreign_key_create_info();
+  char *storage_get_foreign_key_create_info();
+  char *wrapper_get_tablespace_name(THD *thd, char *name, uint name_len);
+  char *storage_get_tablespace_name(THD *thd, char *name, uint name_len);
+  bool wrapper_can_switch_engines();
+  bool storage_can_switch_engines();
+  int wrapper_get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+  int storage_get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+#ifdef MRN_HANDLER_HAVE_GET_PARENT_FOREIGN_KEY_LIST
+  int wrapper_get_parent_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+  int storage_get_parent_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+#endif
+  uint wrapper_referenced_by_foreign_key();
+  uint storage_referenced_by_foreign_key();
+  void wrapper_init_table_handle_for_HANDLER();
+  void storage_init_table_handle_for_HANDLER();
+  void wrapper_free_foreign_key_create_info(char* str);
+  void storage_free_foreign_key_create_info(char* str);
 };
 
 #ifdef __cplusplus
